import numpy as np
import pandas as pd
from .SurvivalTree import SurvivalTree
from .scoring import concordance_index
from joblib import Parallel, delayed
import multiprocessing


class RandomSurvivalForest:
    x = None
    y = None
    bootstrap_idxs = None
    oob_idxs = None
    oob_score = None
    trees = []

    def __init__(self, n_estimators=2, timeline=None, min_leaf=1, unique_deaths=1, n_jobs=None, random_state=None):
        """
        A Random Survival Forest is a tool especially designed for survival analysis.
        :param n_estimators: The numbers of trees in the forest.
        :param timeline: The timeline used for the prediction.
        :param min_leaf: The minimum number of samples required to be at a leaf node. A split point at any depth will
        only be considered if it leaves at least min_leaf training samples in each of the left and right branches.
        :param unique_deaths: The minimum number of unique deaths required to be at a leaf node.
        :param n_jobs: The number of jobs to run in parallel for fit. None means 1.
        """
        self.n_estimators = n_estimators
        self.min_leaf = min_leaf
        self.timeline = timeline
        self.unique_deaths = unique_deaths
        self.n_jobs = n_jobs
        self.random_state = random_state

    def fit(self, x, y):
        """
        Build a forest of trees from the training set (X, y).
        :param x: The input samples. Should be a Dataframe with the shape [n_samples, n_features].
        :param y: The target values as a Dataframe with the survival time in the first column and the event
        in the second with the shape [n_samples, 2]
        :return: self: object
        """

        self.x = x
        self.y = y
        if self.n_jobs == -1:
            self.n_jobs = multiprocessing.cpu_count()
        elif self.n_jobs is None:
            self.n_jobs = 1
        self.bootstrap_idxs = self.draw_bootstrap_samples(x)
        self.trees = Parallel(n_jobs=self.n_jobs)(delayed(self.create_tree)(bootstrap_idx)
                                                  for bootstrap_idx in self.bootstrap_idxs)
        self.oob_score = self.compute_oob_score

        return self

    def create_tree(self, bootstrap_idx):
        """
        Grows a survival tree for the bootstrap samples.
        :param bootstrap_idx: Indices of the bootstrap samples.
        :return: SurvivalTree
        """
        n_features = int(round(np.sqrt(self.x.shape[1]), 0))
        if self.random_state is None:
            f_idxs = np.random.permutation(self.x.shape[1])[:n_features]
        else:
            f_idxs = np.random.RandomState(seed=self.random_state).permutation(self.x.shape[1])[:n_features]
        tree = SurvivalTree(self.x.iloc[bootstrap_idx, :], self.y.iloc[bootstrap_idx, :],
                            f_idxs=f_idxs, n_features=n_features, timeline=self.timeline,
                            unique_deaths=self.unique_deaths, min_leaf=self.min_leaf, random_state=self.random_state)
        return tree

    def compute_oob_ensembles(self):
        """
        Compute OOB ensembles.
        :return: List of oob ensemble for each sample.
        """
        oob_ensemble_chfs = []
        for sample_idx in range(self.x.shape[0]):
            denominator = 0
            numerator = 0
            for b in range(self.n_estimators):
                if sample_idx not in self.bootstrap_idxs[b]:
                    sample = self.x.iloc[sample_idx].to_list()
                    chf = self.trees[b].predict(sample)
                    denominator = denominator + 1
                    numerator = numerator + 1 * chf

            if denominator == 0:
                continue
            else:
                ensemble_chf = numerator/denominator
                oob_ensemble_chfs.append(ensemble_chf)
        return oob_ensemble_chfs

    def compute_oob_score(self):
        """
        Compute the oob score (concordance-index).
        :return: c-index of oob samples
        """
        oob_ensembles = self.compute_oob_ensembles()
        c = concordance_index(y_time=self.y.iloc[:, 0], y_pred=oob_ensembles, y_event=self.y.iloc[:, 1])
        return c

    def predict(self, xs):
        """
        Predict survival for xs.
        :param xs:The input samples
        :return: List of the predicted cumulative hazard functions.
        """
        preds = []
        for x in xs.values:
            chfs = []
            for q in range(self.n_estimators):
                chfs.append(self.trees[q].predict(x))
            preds.append(pd.concat(chfs).groupby(level=0).mean())
        return preds

    def draw_bootstrap_samples(self, data):
        """
        Draw bootstrap samples
        :param data: Data to draw bootstrap samples of.
        :return: Bootstrap indices for each of the trees
        """
        bootstrap_idxs = []
        for i in range(self.n_estimators):
            no_samples = len(data)
            data_rows = list(data.index)
            if self.random_state is None:
                bootstrap_idx = np.random.choice(data_rows, no_samples)
            else:
                np.random.seed(self.random_state + i)
                bootstrap_idx = np.random.choice(data_rows, no_samples)

            bootstrap_idxs.append(bootstrap_idx)

        return bootstrap_idxs




import numpy as np
from .Node import Node
from . import splitting


class SurvivalTree:

    def __init__(self, x, y, f_idxs, n_features, timeline, unique_deaths=1, min_leaf=1, random_state=None):
        """
        A Survival Tree to predict survival.
        :param x: The input samples. Should be a Dataframe with the shape [n_samples, n_features].
        :param y: The target values as a Dataframe with the survival time in the first column and the event.
        :param f_idxs: The indices of the features to use.
        :param n_features: The number of features to use.
        :param timeline: The timeline used for the prediction.
        :param unique_deaths: The minimum number of unique deaths required to be at a leaf node.
        :param min_leaf: The minimum number of samples required to be at a leaf node. A split point at any depth will
        only be considered if it leaves at least min_leaf training samples in each of the left and right branches.
        """
        self.x = x
        self.y = y
        self.f_idxs = f_idxs
        self.n_features = n_features
        self.timeline = timeline
        self.min_leaf = min_leaf
        self.unique_deaths = unique_deaths
        self.random_state = random_state
        self.score = 0
        self.index = 0
        self.split_val = None
        self.split_var = None
        self.lhs = None
        self.rhs = None
        self.chf = None
        self.grow_tree()

    def grow_tree(self):
        """
        Grow the survival tree recursively as nodes.
        :return: self
        """
        self.score, self.split_val, self.split_var, lhs_idxs_opt, rhs_idxs_opt = splitting.find_split(self)

        if self.random_state is None:
            lf_idxs = np.random.permutation(self.x.shape[1])[:self.n_features]
            rf_idxs = np.random.permutation(self.x.shape[1])[:self.n_features]
        else:
            lf_idxs = np.random.RandomState(seed=self.random_state).permutation(self.x.shape[1])[:self.n_features]
            rf_idxs = np.random.RandomState(seed=self.random_state).permutation(self.x.shape[1])[:self.n_features]

        self.lhs = Node(x=self.x.iloc[lhs_idxs_opt, :], y=self.y.iloc[lhs_idxs_opt, :],
                        tree=self, f_idxs=lf_idxs, n_features=self.n_features,
                        unique_deaths=self.unique_deaths, min_leaf=self.min_leaf, random_state=self.random_state)

        self.rhs = Node(x=self.x.iloc[rhs_idxs_opt, :], y=self.y.iloc[rhs_idxs_opt, :],
                        tree=self, f_idxs=rf_idxs, n_features=self.n_features,
                        unique_deaths=self.unique_deaths, min_leaf=self.min_leaf, random_state=self.random_state)

        return self

    def predict(self, x):
        """
        Predict survival for x.
        :param x: The input sample.
        :return: The predicted cumulative hazard function.
        """
        if x[self.split_var] <= self.split_val:
            self.lhs.predict(x)
        else:
            self.rhs.predict(x)
        return self.chf



from setuptools import setup

with open('README.md') as f:
    readme = f.read()

setup(
  name = 'random_survival_forest',         # How you named your package folder (MyLib)
  packages = ['random_survival_forest'],   # Chose the same as "name"
  version = '0.4',      # Start with a small number and increase it with every change you make
  license="MIT License",        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  long_description=readme,
  long_description_content_type="text/markdown",
  description = 'A Random Survival Forest implementation inspired by Ishwaran et al.',   # Give a short description about your library
  author = 'Julian Späth',                   # Type in your name
  author_email = 'spaethju@posteo.de',      # Type in your E-Mail
  url = 'https://github.com/julianspaeth/random-survival-forest',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/julianspaeth/random-survival-forest/archive/v0.1-alpha.tar.gz',    # I explain this later on
  keywords = ['survival-analysis', 'survival-prediction', 'machine-learning', 'random-forest', 'random-survival-forest'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'numpy',
          'pandas',
          'joblib',
          'multiprocess',
          'lifelines'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package    'Intended Audience :: Developers',      # Define that your audience are developers
    'Topic :: Software Development :: Build Tools',    'License :: OSI Approved :: MIT License',   # Again, pick a license    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
  ],
)



import numpy as np
from lifelines import NelsonAalenFitter
from . import splitting


class Node:

    score = 0
    split_val = None
    split_var = None
    lhs = None
    rhs = None
    chf = None
    chf_terminal = None
    terminal = False

    def __init__(self, x, y, tree, f_idxs, n_features, unique_deaths=1, min_leaf=1, random_state=None):
        """
        A Node of the Survival Tree.
        :param x: The input samples. Should be a Dataframe with the shape [n_samples, n_features].
        :param y: The target values as a Dataframe with the survival time in the first column and the event.
        :param tree: The corresponding Survival Tree
        :param f_idxs: The indices of the features to use.
        :param n_features: The number of features to use.
        :param unique_deaths: The minimum number of unique deaths required to be at a leaf node.
        :param min_leaf: The minimum number of samples required to be at a leaf node. A split point at any depth will
        only be considered if it leaves at least min_leaf training samples in each of the left and right branches.
        """
        self.x = x
        self.y = y
        self.tree = tree
        self.f_idxs = f_idxs
        self.n_features = n_features
        self.unique_deaths = unique_deaths
        self.random_state = random_state
        self.min_leaf = min_leaf
        self.grow_tree()

    def grow_tree(self):
        """
        Grow tree by calculating the Nodes recursively.
        :return: self
        """
        unique_deaths = self.y.iloc[:, 1].reset_index().drop_duplicates().sum()[1]

        if unique_deaths <= self.unique_deaths:
            self.compute_terminal_node()
            return self

        self.score, self.split_val, self.split_var, lhs_idxs_opt, rhs_idxs_opt = splitting.find_split(self)

        if self.split_var is None:
            self.compute_terminal_node()
            return self

        if self.random_state is None:
            lf_idxs = np.random.permutation(self.x.shape[1])[:self.n_features]
            rf_idxs = np.random.permutation(self.x.shape[1])[:self.n_features]
        else:
            lf_idxs = np.random.RandomState(seed=self.random_state).permutation(self.x.shape[1])[:self.n_features]
            rf_idxs = np.random.RandomState(seed=self.random_state).permutation(self.x.shape[1])[:self.n_features]

        self.lhs = Node(self.x.iloc[lhs_idxs_opt, :], self.y.iloc[lhs_idxs_opt, :], self.tree,
                        lf_idxs, self.n_features, min_leaf=self.min_leaf, random_state=self.random_state)

        self.rhs = Node(self.x.iloc[rhs_idxs_opt, :], self.y.iloc[rhs_idxs_opt, :], self.tree,
                        rf_idxs, self.n_features, min_leaf=self.min_leaf, random_state=self.random_state)

        return self

    def compute_terminal_node(self):
        """
        Compute the terminal node if condition has reached.
        :return: self
        """
        self.terminal = True
        self.chf = NelsonAalenFitter()
        t = self.y.iloc[:, 0]
        e = self.y.iloc[:, 1]
        self.chf.fit(t, event_observed=e, timeline=self.tree.timeline)

        return self

    def predict(self, x):
        """
        Predict the cumulative hazard function if its a terminal node. If not walk through the tree.
        :param x: The input sample.
        :return: Predicted cumulative hazard function if terminal node
        """
        if self.terminal:
            self.tree.chf = self.chf.cumulative_hazard_
            self.tree.chf = self.tree.chf.iloc[:, 0]
            return self.tree.chf

        else:
            if x[self.split_var] <= self.split_val:
                self.lhs.predict(x)
            else:
                self.rhs.predict(x)
